
!
!  Module      : ogprep_mod
!
! Authors     : Vladimir Yershov (MSSL)
!          
!
! Description : Alignes OM detector (grism) spectra along y-coordinate
! 
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
! 
! History : Created  : V0.1 October 2002 
! VNY 06.11.2003 Parameter for the output undistorted image made non-mandatory
! VNY 07.11.2003 The background around the rotated image is set to -1
! VNY 11.11.2003 the code is tiden up
! 25.11.2003 (VNY) distortion correction improvement
! 07.05.2004 (VNY) trying to recover the lost pixels at the edge of the image
! Temporarily restoring the wrong algorithm of computing
! the binning factors (to make it compatible with the imaging tasks
! 11.05.2004 (VNY) making pixel numbering consistent before and after adding
!  new pixels
! 13.05.2004 (VNY) -updating and closing SSC-SPR-3318 on the missing pixels
! at the edge of the output image
! 01.07.2004 (VNY) - fixing SSC-SPR-3346: the extra pixels added at the 
! edges of the window have to be binned (for the consistency reasons)
! 26.07.2004 (VNY) removing WCS-coordinate system which leads to an odd 
! view of the coordinates in the image (SSC-SPR-3363)
! 07.09.2004 VNY Rotation angles of the grism image are changed
! UV: -6.3 -> -6.4   VIS: 5.6 -> 5.5
! 08.09.2004 VNY The rotation angles have been returned to their 
! initial values because there is no apparent improvement from the 
! new ones
! 01.06.2005 VNY Corrected the problem of the moire pattern by averaging the 
! pixels (in vertical direction) around the pattern-fringe. Fixing SSC-SPR-3462
! 02.06.2006 VNY Introducing a flag that would indicate the extractability of 
! the spectrum from the grism-image (SPR-3606)
! 26.06.2006 VNY adjusting the RA and DEC values in the header for the reference pixel
! defined by the grism-default reference
! 27.06.2006 VNY Computing the distortion correction for the grism reference pixel
! 30.06.2006 VNY Introducing a new keyword (OMGWIN) into the header of the
! output file (rotated image). The value 1 would mean that the window size is
! acceptable for the spectrum extraction, 0 - too small. This is done as an
! alternative to the return flag, which is not consistent with the rules for 
! the SAS tasks (SPR-3616) 
! 11.07.2006 VNY correcting the values in the MODES table of the output file
! (previously they were copied from the input file) 
! 2.08.2006 VNY Checking the spectra extractability value and fixing a bug:
! the x and y-coordinates of the default target were converted into unbinned 
! values, but this is not correct because the default target is given in the 
! binned coordinates
! 18.08.06 Checking the image rotation without undistorting it
! 07.09.2006 Added two new keywords to the header of the rotated image:
!            OMGXMIN and OMGYMIN in order to pass the information about the 
!            rotation to the OMATT task
! 21.09.2006 VNY Modifying the DETSEC keyword of the rotated image: it should
! be the same as in the unrotated image in order to be able to process correctly
! the RA and DECs of the sources by the task omgrism. Also changing the 
! corresponding values in the MODES table of the rotated image
! 20.10.2006 VNY modifying the CRPIX1 and CRPIX2 of the undistorted and rotated
!            images to make it based on the (1024.5,1025.5)-pixel. Also modifying
!            the DETSEC keyword of the undistorted image (it should be the same
!            as of the rotated image)
! 02.11.2006 VNY Modifying the CRPIX1 and CRPIX2 of the undistorted and rotated 
!            images so that to reduce the original (1024.5,1024.5)-value to
!            (36,105), which is close to the target and which is obtained 
!            by comparing the RA, DEC with the sky-map
! 09.11.2006 VNY Cleaning up from the test messages, preparing for the update
!            according to SPR-3633
! 25.01.2007 VNY Fixed the bag, which was shifting the wavelength scale
!            by about 50 Angstroems (subroutine UndistortImage3/UndistortImage4)
!*************************************************************************** 

module omgprep_mod

  use param
  use oal
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod
  use allocation_mod 
  use utility_mod
  use omconstants_mod
!  use omdistortion_mod

  use types

implicit none
 
  private

   public omgprep    

contains

!---------------------------------------------------------------------
subroutine omgprep
!---------------------------------------------------------------------

    type(DataSetT) :: inDataSetH                                       ! Input dataset handle.
!    type(DataSetT) :: sourceListDataSetH                              ! Source List dataset handle.
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.
!    type(DataSetT) :: smoothDataSetH                                  ! hadle for the smoothed output dataset
    type(DataSetT) :: undistDataSetH                                  ! hadle for the undistorted output dataset
    type(ArrayT) :: inArrayH                                          ! Input array handle.
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.
    type(ArrayT) :: undistArrayH                                      ! smooth array handle
    type(AttributableT) :: sourceH                                    ! Source attributable handle.
    type(AttributableT) :: objectH                                    ! Object attributable handle.
    type(BlockT) :: blockH                                            ! Block handle.

    type (SpacecraftAttitudeType) attitude

    character(len=255) :: set                                         ! OM image intermediate filename
    character(len=255) :: PPSOSWSet                                   ! PPS OM OSW FITS sky image filename (rotated).
    character(len=255) :: undistSet                                   ! filename for the undistorted image    
    character(len=20):: rastring, decstring   ! strings for RA and DEC
    character(len=1) :: cs                    ! sign of dec +/-
    character(len=10)           :: datamode   ! datamode attribute
    character(len=6)            :: omFilter   ! the OM filter
    character(len=6)            :: tempFilter   ! temporary the OM filter

    integer                    :: dim(2), nx, ny     ! size of the image
    integer                    :: xBinning
    integer                    :: yBinning
    integer(kind=int32) :: xbin2,ybin2
    integer             :: nxout, nyout        ! size of the output image
    integer             :: nxdist, nydist        ! size of the output udistorted image
    integer :: i, j, k                           ! counters
    integer             :: xbin, ybin         ! binning factors
    integer             :: xsize, ysize       ! size of the image
    integer             :: x0, y0             ! offset of the image
    integer             :: idmsf(4), ihmsf(4)  
    integer                     :: OM_MODE    ! the om mode

    real(kind=single), pointer :: oswImage(:, :)
    real                       :: refDX, refDXext
    real                       :: refDY, refDYext
    real                       :: rotatedRefDX
    real                       :: rotatedRefDY
    real                       :: refDXcorrection
    real                       :: refDYcorrection
    real(kind=double)  :: itrans(6)     ! its inverse

    ! VNY 07.05.2004 expanded image to allow change the shape of the undistorted image
    real(kind=single), pointer  :: oswImageExt(:,:)       ! The expanded osw Image
    integer :: nxExt, nyExt

    ! sky coordinate image variables
    real(kind=single), pointer  :: outImage(:,:)       ! The output image
    real(kind=single), pointer :: undistImageP(:,:)     ! The undistorted output image
    real(kind=single), dimension(:, :), allocatable :: undistImage 
    real(kind=single)          :: exposureTime         ! The exposure time for an input image
    real(kind=double)  :: trans(6) ! rotation transformation matrix
                         
    real(kind=double)  :: rotationCorrection = 90.0_double! Correction for coordinate error
    real(kind=double)  :: dXbinning, dYbinning
    real(kind=double)  :: sinRollAngle, cosRollAngle
   
    real(kind=single)  :: refx, refy      ! SRR added 4/10/00
    real(kind=double) ::  refxa, refya   
    real(kind=double)  :: xmax,xmin,ymax,ymin            
    real(kind=double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft 
    real(kind=double) :: ra_scx2, dec_scx2, pa_scx2  ! The ra, dec and pa of the spacecraft  
    real(kind=double) :: ra_scx3, dec_scx3   

    real(kind=double) :: ra_point, dec_point        ! position of the OM tangent point
    real(kind=double) :: rollangle                 ! the roll angle
    real(kind=double) :: time                        
    real(kind=double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(kind=double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(kind=double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(kind=double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(kind=double) :: pointvec(3)        ! s/c pointing
    real(kind=double) :: pointvec2(3)       
    real(kind=double) :: euler_axis(3)              ! euler axis for rotation
    real(kind=double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(kind=double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(kind=double) :: omvec(3)                   ! om pointing vector
    real(kind=double) :: rotmat(3, 3)               ! rotation matrix
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real, pointer     :: sigmax(:), sigmay(:)       ! distortion sigmas
    real(kind=double) :: r1                         ! modulus
    real(kind=double) :: currentEpoch
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double)           :: lambda           ! wavelength for CAL
    real(kind=double)           :: plateScale       ! the platescale
    real :: white, black!, grey, grey2
    real(kind=single) :: nullValue
   integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1 !+   ! Coordinates of the OSW origin
   integer(kind=int32) :: windowDX, windowDY !+
   integer(kind=int32) :: winX0, winY0, winX1, winY1, winDX, winDY
   integer(kind=int32) :: winX0ext, winY0ext, winX1ext, winY1ext
   integer(kind=int32) :: winX0rot, winY0rot, winX1rot, winY1rot
   character(len=1) :: ch1, ch2, ch3, ch4, ch5
   character(len=255) :: window0 

    logical(kind=Bool) :: binBPE                    ! BPE binning on?
    logical(kind=bool) :: produceUndistortedImage ! condition for producin an auxiliar output image
!    integer                       :: istat              ! status for memory allocation
    integer(kind=int32)           :: filterIdInt
    real(kind=single) :: averImage
    real(kind=single) :: sum !, cosPA, sinPA
    character(len=255) :: detsec, detsize
    integer(kind=int8), dimension(:, :), pointer :: qualArrayP        ! Quality array pointer.
    integer(kind=int8), dimension(:, :), pointer :: outQualityArrayP  ! Quality-2 array pointer.
    logical(kind=bool), dimension(:, :), allocatable :: qualArray     ! Quality array.
    integer(kind=int8), dimension(:, :), allocatable :: outQualArray 
!--- VNY 11.05.2004
    integer(kind=int8), dimension(:, :), allocatable :: extQualArray 
    integer(kind=int8), dimension(:, :), pointer :: extQualityArrayP  ! extended Quality array pointer.
    type(ArrayT) :: extQualityArrayH
!---
    logical(kind=bool) :: hasQUALITY
    type(ArrayT) :: outQualityArrayH
    integer(kind=int32) :: allocErr                                   ! Allocate error status.
    character(len=255) :: attrValue
    integer(kind=int32) :: int32Value
    real(kind=single)  :: real32Value
    integer(kind=int32) :: calcMode1
    integer :: extractability
    character(len=10) :: primInstrValue
!    real(kind=double) :: xOrigin, yOrigin, xDistorted, yDistorted
    real(double) :: grismRollangle             ! the grism filter roll angle    
    real(double) :: initialRollangle 
    integer(kind=int32) :: xTarget, yTarget
! 
    type(TableT) :: modesTableH                                       ! Modes table dataset handle.
    integer(int32) :: modeIMG
    integer(int32), dimension(1) :: columnDim                    ! Column dimension.
    integer(int32) :: windowX0a                                   ! Keywords.
    integer(int32) :: windowY0a                                   ! Keywords.
    integer(int32) :: windowDXa                                   ! Keywords.
    integer(int32) :: windowDYa                                   ! Keywords.
!
    real(kind=double) :: testX, testY, testX1, testY1

    nullValue = real(-99999)
    produceUndistortedImage = .false.
    !*************************************
    ! Get the command-line parameters
    !*************************************
    set = stringParameter("set")
    PPSOSWSet = stringParameter("outset")
    !smoothSet = stringParameter("smoothset")
    undistSet = stringParameter("undistset")
    if (undistSet .ne. "") then
       produceUndistortedImage = .true.
    endif

!******************************************** 

    ! the image contains information about the science window
    inDataSetH = DataSet(set, modify) 
    inArrayH = array(inDataSetH, "PRIMARY", modify)
    exposureTime = real32Attribute(inDataSetH, "EXPOSURE") ! Image exposure time
!=================================

    hasQUALITY= .false.
    if (hasBlock(inDataSetH, "QUALITY")) then
       hasQUALITY= .true.
    endif

    xbin = int32Attribute(inDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(inDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    binBPE = booleanAttribute(inDataSetH, "BINBPE")

    ! Calculate image binning.
! Temporarily restoring the wrong algorithm of computing
! the binning factors (to make it compatible with the imaging tasks
    !if (xbin .eq. 0) xbin=1
    !if (ybin .eq. 0) ybin=1
    !xBinning = xbin
    !yBinning = ybin
    xBinning = 2 ** xbin
    yBinning = 2 ** ybin

    if (binBPE) then
        xBinning = 2.0
        yBinning= 2.0
    end if
    xbin2=xBinning
    ybin2=yBinning

    dXbinning = dble(xBinning)  ! Need the double values of xBinning and 
    dYbinning = dble(yBinning)  ! yBinning fairly frequently

    !-(origin of the OSW)
       if (hasAttribute(inArrayH,"WINDOWX0")) then
          windowX0= int32Attribute(inArrayH, "WINDOWX0")
          windowY0= int32Attribute(inArrayH, "WINDOWY0")
          windowDX= int32Attribute(inArrayH, "WINDOWDX")
          windowDY= int32Attribute(inArrayH, "WINDOWDY")
          winX0=windowX0/xBinning
          winY0=windowY0/yBinning
          winDX=windowDX/xBinning
          winDY=windowDY/yBinning
          winX1=winX0+winDX-1.0
          winY1=winY0+winDY-1.0
       else
          if (hasAttribute(inArrayH,"DETSEC")) then
             window0 = stringAttribute(inArrayH, "DETSEC")
             read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
             winX0=windowX0
             winY0=windowY0
             winX1=windowX1
             winY1=windowY1
             winDX=windowX1-windowX0+1
             winDY=windowY1-windowY0+1
          else
             ! Check if there is DETSEC in the principal header
             if (hasAttribute(indataSetH,"DETSEC")) then
                window0 = stringAttribute(inDataSetH, "DETSEC")
                read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                     ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
                winX0=windowX0
                winY0=windowY0
                winX1=windowX1
                winY1=windowY1
                winDX=windowX1-windowX0+1
                winDY=windowY1-windowY0+1
             else
                dim = dimensions(inArrayH)   ! Get the size of the image 
                winX0=0
                winY0=0
                winDX = dim(1)
                winDY = dim(2)
                winX1= winDX-1.0
                winY1= winDY-1.0
             endif ! hasAttribute(inDataSetH, DETSEC...
          endif ! hasAttribute(inArrayH, DETSEC...
       endif ! hasAttribute(inArrayH, WINDOWX0...

    tempFilter="UVW1"   
    omFilter = stringAttribute(inDataSetH, "FILTER")
    if (trim(omFilter) .eq. "Grism1") then 
       omFilter="GRISM1"
       tempFilter="UVW1"
    endif
    if (trim(omFilter) .eq. "Grism2") then
       omFilter="GRISM2"
       tempFilter="V"
    endif

    if (omFilter .eq. "GRISM2  ") grismRollangle=-5.6
    if (omFilter .eq. "GRISM1  ") grismRollangle=6.3

    if (trim(dataMode) .eq. "IMAGING") then 
       OM_MODE = IMAGE
    else
       OM_MODE = FAST
    end if

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSetH, x0, y0, xsize, ysize)

    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)   

    allocate (qualArray(0 : nx-1, 0 : ny-1), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")

    ! VNY 07.05.2004 Allocate the expanded image array
    ! VNY 01.07.2004 Taking into account the binning factor for these extra pixels
    nxExt=nx+20/xBinning
    nyExt=ny+40/yBinning
    winX0ext=winX0-10/xBinning
    winX1ext=winX1+10/xBinning
    winY0ext=winY0-20/yBinning
    winY1ext=winY1+20/yBinning
!
    allocate (oswImageExt(0:nxExt-1, 0:nyExt-1), stat=allocErr)
    call CheckAllocation(allocErr, "oswImageExt")
!    allocate (qualArrayExt(0 : nxExt-1, 0 : nyExt-1), stat = allocErr)
!    call CheckAllocation(allocErr, "qualArrayExt")

    ! Assign the values of the extended array
    oswImageExt=-1.0
    do i=0,nx-1
       do j=0,ny-1
          oswImageExt(i+10/xBinning,j+20/yBinning)=oswImage(i,j)
       enddo
    enddo    

    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArray=.false.
    if (hasQUALITY) then
       qualArrayH = array(inDataSetH, "QUALITY", read)
       qualArrayP => int8Array2Data(qualArrayH)
       !outQualArray=qualArrayP
       where (qualArrayP .ne. 0) 
          qualArray = .true.
       elsewhere
          qualArray = .false.
       end where
    endif

    !***********************************************************
    ! Check the source shapes 
    !***********************************************************
    black = minval(oswImage)
    white = maxval(oswImage)
    averImage=0.0
    sum=0.0

    do i=0,nx-1
       do j=0,ny-1
          averImage=averImage+oswImage(i,j)
          sum=sum+1.0
       enddo
    enddo

    if (sum .gt. 0.0) then
       averImage=averImage/sum
    endif

    !*******************************************************
    ! get the astrometry info keywords 
    !*******************************************************
    ra_scx = real64Attribute(inDataSetH, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(inDataSetH, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(inDataSetH, "PA")        ! Spacecraft PA in degrees

    initialRollangle=pa_scx+grismRollangle
    !********************************************
    ! Transform the source position angles
    ! to the equatorial coordinate system
    !*******************************************
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    !************************************
    ! Get the filter Id integer value
    !************************************
    filterIdInt = filterIdToInt(omFilter)
    !*********************
    ! Set the CAL state
    !*********************
    ! VNY 11.12.2003 The following line is to fool CAL, which has a bug somwhere
    ! and which does not compute the polinomials for the grism filters
    filterIdInt=filterIdToInt(tempFilter)

    call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)
    call CAL_omGetPlateScale(lambda, plateScale)  ! Get the plate-scale (arcsecs per pixel) 
    !plateScale=0.476513
   
!    call InitialiseDistortion(winX0, winY0, xBinning, yBinning)

    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(inDataSetH)

!    currentEpoch = real32Attribute(sourceListDataSetH, "EQUINOX")
!***
    currentEpoch=2000.0
!***
    !*********************************************************
    ! Need the current epoch for the astrometry correction
    !*********************************************************
    currentEpoch = 1998.0_double + time / (365.25_double * 86400.0_double)
    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)
    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    rollangle=0.0

    ! Visual grism rollangle (Grism2)
    if (omFilter .eq. "Grism2  " .or. omFilter .eq. "GRISM2  ") rollangle=-5.6
    if (omFilter .eq. "Grism1 " .or. omFilter .eq. "GRISM1  ") rollangle=6.3

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)

    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

    !**********************************************
    ! rotationCorrection corrects for OM axes definition.
    !**********************************************
    rotationCorrection =0.0
    !rotationCorrection = rotationCorrection * degreesToRadians 

    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the modulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !-------
    euler_axis = -rollangle * pointvec        !leave in (SRR)

    ! convert euler axis vector to a rotation matrix
    call sla_dav2m(euler_axis, rotmat) 

    ! initialise output vector
    omvec = 0.0_double

!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************
    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)
    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************
    ra_point=0.0
    dec_point=0.0
    call sla_dcc2s(omvec, ra_point, dec_point)

    ! Protection against the NaN values of ra ad dec
    if (ra_point .eq. nullValue .and. dec_point .eq. nullValue) then
       ra_point=0.0
       dec_point=0.0
    endif
    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)

    deltax => AllocateSingle(0, 0, "deltax")
    deltay => AllocateSingle(0, 0, "deltay")
    sigmax => AllocateSingle(0, 0, "sigmax")
    sigmay => AllocateSingle(0, 0, "sigmay")
 

    !****************************************************************
    !  NOW CREATE THE OUTPUT PPS PRODUCT OM OSW FITS IMAGE 
    !  This is done by resampling the detector image on a sky grid. 
    ! the transformation is currently a simple rotation
    ! Define the transformation terms (note the offsets will change)
    ! The image is rotated to align the spectrum with the vertical direction
    !****************************************************************

    rollangle=rollangle * degreesToRadians 
    trans(1) =  0.0_double
    trans(2) =  dcos(-rollangle)
    trans(3) =  dsin(-rollangle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollangle)
    trans(6) =  dcos(-rollangle)
    sinRollangle = dsin(rollangle)
    cosRollangle = dcos(rollangle)
    !*******************************************
    ! create the output source file
    !*******************************************

    PPSOSWDataSetH = DataSet(trim(PPSOSWSet), create)
    !smoothDataSetH = DataSet(trim(smoothSet), create)
    if (produceUndistortedImage) undistDataSetH = DataSet(trim(undistSet), create)

!  determine the image boundaries for the undistorted image
    itrans(1)=0.0_double
    itrans(2)=1.0_double
    itrans(3)=0.0_double    
    itrans(4)=0.0_double
    itrans(5)=0.0_double
    itrans(6)=1.0_double
! (the rotation matrix itrans is set to the zero rotation angle)

    nxdist=nxExt
    nydist=nyExt


!---
!**************************************************************************
!  Test creating the undistorted image (21.07.03)
!**************************************************************************
    dim(1) = nxdist     ! The dimensions of the output image
    dim(2) = nydist

    ! copy the header across
    !
    if (produceUndistortedImage) then 
       !call copyAllAttributes(attributable(inDataSetH), attributable(undistDataSetH))
       undistArrayH = addArray(undistDataSetH, "PRIMARY",  Real32, dim, "Counts", "IMAGE")

       !sourceH = attributable(inDataSetH)
       objectH=attributable(undistArrayH)
       !call addAttributes(objectH, sourceH)

       call setAttribute(undistDataSetH, "FRAMTIME", 21)
       ! add content keyword
       call setAttribute(undistDataSetH, "CONTENT", "OM OSW FITS UNDISTORTED IMAGE", &
            "Content of FITS file")
       call setAttribute(undistDataSetH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")

       blockH=block(inDataSetH, "PRIMARY", read)
       sourceH=attributable(blockH)

       !call addAttributes(objectH, sourceH)

       call setAttribute(objectH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
    
       ! 11.05.2004 Correct the quality array
       !blockH=block(inDataSetH, "QUALITY", read)
       !call copyBlock(undistDataSetH, blockH)

       extQualityArrayH = addArray(undistDataSetH, "QUALITY", integer8, dim, "flags", "IMAGE")
       allocate (extQualArray(0 : nxdist-1, 0 : nydist-1), stat = allocErr)
       call CheckAllocation(allocErr, "extQualArray")
       extQualArray=0

       !blockH=block(inDataSetH, "MODES", read)
       !call copyBlock(undistDataSetH, blockH)
    endif ! produceUndistortedImage



    
    !**********************************************************************
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !*********************************************************************

    primInstrValue=''
    call GetPrimaryInstrument(PPSOSWSet,primInstrValue)

    call  GetImageBoundaries(nxExt, nyExt, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax, refxa, refya, omFilter, &
                             primInstrValue, binBpe, xTarget, yTarget)
    refxa=xTarget
    refya=yTarget
    ! invert transformation matrix for later use
    call invert_trans(trans, itrans)    ! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin

    ! VNY 11.05.2004 limits of the rotated image that should be put 
    ! into DETSEC

    winX0rot=winX0ext+xMin
    winX1rot=winX0rot+nxout
    winY0rot=winY0ext+yMin
    winY1rot=winY0rot+nyout

    !***********************************************************************************
    ! set up the output image with the appropriate size (xmax,ymax already referenced to 
    ! origin of containing (output) box so these equal the output box dimensions).
    !**********************************************************************************
    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout

    ! copy the header across
    PPSOSWArrayH    = addArray(PPSOSWDataSetH, "PRIMARY",  Real32, dim, "Counts", "IMAGE")  ! For the image array

    sourceH = attributable(inDataSetH)
    objectH=attributable(PPSOSWArrayH)
   
    !call addAttributes(objectH, sourceH)
    ! add content keyword
    call setAttribute(PPSOSWDataSetH, "CONTENT", "OM OSW FITS ROTATED IMAGE", &
         "Content of FITS file")
    ! Add attribute HDUCLASS
    call setAttribute(PPSOSWDataSetH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")

    blockH=block(inDataSetH, "PRIMARY", read)
    sourceH=attributable(blockH)
    !call addAttributes(objectH, sourceH)
    call setAttribute(objectH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
    write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",winX0rot, ":", winX1rot-1, &
         & ",",winY0rot, ":", winY1rot-1, "]" 
!****************************************************************************
! Setting detsec to be the sabe as of the extended unrotated image
!*****************************************************************************
    window0 = stringAttribute(inDataSetH, "DETSEC")
    read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
         ch1,winX0, ch2, winX1, ch3,  winY0, ch4, winY1, ch5
    write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",winX0-10/xBinning, ":", winX1+10/xBinning, &
         & ",",winY0-20/yBinning, ":", winY1+20/yBinning, "]" 
!*****************************************************************************
    if (xbin .ne. 0. .and. ybin .ne. 0.) then
       write (detsize,'(a6,i4,a6,i4,a1)') "[   1:", INT(2048.0/real(xbin)),",   1:",INT(2048.0/real(ybin)), "]"
    else
       write (detsize,'(a21)') "[   1:2048,   1:2048]"
    endif

    ! Set the principal attributes for the output data set
    if (hasAttribute(inDataSetH, "EXTVER")) then
       int32Value=int32Attribute(inDataSetH, "EXTVER")
       call setAttribute(PPSOSWDataSetH, "EXTVER", int32Value, "Extension Version" )
    endif
    !
    if (hasAttribute(inDataSetH, "TELESCOP")) then
       attrValue=stringAttribute(inDataSetH, "TELESCOP")
       call setAttribute(PPSOSWDataSetH, "TELESCOP", trim(attrValue), "XMM Mission" )
    endif
    !
    if (hasAttribute(inDataSetH, "INSTRUME")) then
       attrValue=stringAttribute(inDataSetH, "INSTRUME")
       call setAttribute(PPSOSWDataSetH, "INSTRUME", trim(attrValue), "Instrument name" )
    endif
    !
    if (hasAttribute(inDataSetH, "CHAIN")) then
       int32Value=int32Attribute(inDataSetH, "CHAIN")
       call setAttribute(PPSOSWDataSetH, "CHAIN", int32Value, "Telemetry chain" )
    endif
    !
    !*****************************************************************************
    ! VNY 28.09.2005 adding the missed attributes
    !
    if (hasAttribute(inDataSetH, "DETNAM")) then
       attrValue=stringAttribute(inDataSetH, "DETNAM")
       call setAttribute(PPSOSWDataSetH, "DETNAM", trim(attrValue), "PRIME or REDUNDANT" )
    endif
    !
    if (hasAttribute(inDataSetH, "OBS_MODE")) then
       attrValue=stringAttribute(inDataSetH, "OBS_MODE")
       call setAttribute(PPSOSWDataSetH, "OBS_MODE", trim(attrValue), "Observation mode (pointing or slew)" )
    endif
    !
    if (hasAttribute(inDataSetH, "ORIGIN")) then
       attrValue=stringAttribute(inDataSetH, "ORIGIN")
       call setAttribute(PPSOSWDataSetH, "ORIGIN", trim(attrValue), "Origin of FITS file" )
    endif
    !
    if (hasAttribute(inDataSetH, "OBJECT")) then
       attrValue=stringAttribute(inDataSetH, "OBJECT")
       call setAttribute(PPSOSWDataSetH, "OBJECT", trim(attrValue), "Name of the observed object" )
    endif
    !
    if (hasAttribute(inDataSetH, "OBSERVER")) then
       attrValue=stringAttribute(inDataSetH, "OBSERVER")
       call setAttribute(PPSOSWDataSetH, "OBSERVER", trim(attrValue), "Name of PI" )
    endif
    !
    ra_scx3=0.0
    if (hasAttribute(inDataSetH, "RA_SCX")) then
       ra_scx2 = real64Attribute(inDataSetH, "RA_SCX")
       ra_scx3=ra_scx2
       call setAttribute(PPSOSWDataSetH, "RA_SCX", ra_scx2, "degrees", "RA of spacecraft pointing" )
    endif
    !
    dec_scx3=0.0
    if (hasAttribute(inDataSetH, "DEC_SCX")) then
       dec_scx2 = real64Attribute(inDataSetH, "DEC_SCX")
       dec_scx3=dec_scx2
       call setAttribute(PPSOSWDataSetH, "DEC_SCX", dec_scx2, "degrees", "DEC of spacecraft pointing" )
    endif
    !
    !if (hasAttribute(inDataSetH, "PA")) then
       !pa_scx2 = real64Attribute(inDataSetH, "PA")
        pa_scx2=initialRollangle
       !call setAttribute(PPSOSWDataSetH, "PA", pa_scx2, "[degrees] ROLL angle of spacecraft pointing" )
       call setAttribute(PPSOSWDataSetH, "PA", pa_scx2, "degrees", "ROLL angle of spacecraft pointing" )
    !endif
    !
    if (hasAttribute(inDataSetH, "RADECSYS")) then
       attrValue=stringAttribute(inDataSetH, "RADECSYS")
       call setAttribute(PPSOSWDataSetH, "RADECSYS", trim(attrValue), "Celestial coordinate system" )
    endif
    !
    if (hasAttribute(inDataSetH, "EQUINOX")) then
       int32Value=int32Attribute(inDataSetH, "EQUINOX")
       call setAttribute(PPSOSWDataSetH, "EQUINOX", int32Value, "Equinox of celestial coordinate system" )
    endif
    !
    if (hasAttribute(inDataSetH, "TIMESYS")) then
       attrValue=stringAttribute(inDataSetH, "TIMESYS")
       call setAttribute(PPSOSWDataSetH, "TIMESYS", trim(attrValue), "XMM time is Terrestrial Time)" )
    endif
    !
    if (hasAttribute(inDataSetH, "MJDREF")) then
       pa_scx2 = real64Attribute(inDataSetH, "MJDREF")
       call setAttribute(PPSOSWDataSetH, "MJDREF", pa_scx2, "d", "1998-01-01T00:00:00 (TT) expressed in MJD" )
    endif
    !
    if (hasAttribute(inDataSetH, "TIMEZERO")) then
       int32Value=int32Attribute(inDataSetH, "TIMEZERO")
       call setAttribute(PPSOSWDataSetH, "TIMEZERO", int32Value, "s", "Clock correction (if not zero)" )
    endif
    !
    if (hasAttribute(inDataSetH, "TIMEUNIT")) then
       attrValue=stringAttribute(inDataSetH, "TIMEUNIT")
       call setAttribute(PPSOSWDataSetH, "TIMEUNIT", trim(attrValue), "All times in s unless specified otherwise" )
    endif
    !
    if (hasAttribute(inDataSetH, "CLOCKAPP")) then
       attrValue=stringAttribute(inDataSetH, "CLOCKAPP")
       call setAttribute(PPSOSWDataSetH, "CLOCKAPP", trim(attrValue), "Clock correction applied" )
    endif
    !
    if (hasAttribute(inDataSetH, "TIMEREF")) then
       attrValue=stringAttribute(inDataSetH, "TIMEREF")
       call setAttribute(PPSOSWDataSetH, "TIMEREF", trim(attrValue), "Reference location of photon arrival times" )
    endif
    !
    if (hasAttribute(inDataSetH, "TASSIGN")) then
       attrValue=stringAttribute(inDataSetH, "TASSIGN")
       call setAttribute(PPSOSWDataSetH, "TASSIGN", trim(attrValue), "Location of time assignment" )
    endif
    !
    if (hasAttribute(inDataSetH, "OSW_ID")) then
       attrValue=stringAttribute(inDataSetH, "OSW_ID")
       call setAttribute(PPSOSWDataSetH, "OSW_ID", trim(attrValue), "Identifier for current OSW" )
    endif
    !
    if (hasAttribute(inDataSetH, "MISSPIX")) then
       pa_scx2 = real64Attribute(inDataSetH, "MISSPIX")
       call setAttribute(PPSOSWDataSetH, "MISSPIX", pa_scx2, "Value indicating missing pixel" )
    endif
    !
    if (hasAttribute(inDataSetH, "ATT_SRC")) then
       attrValue=stringAttribute(inDataSetH, "ATT_SRC")
       call setAttribute(PPSOSWDataSetH, "ATT_SRC", trim(attrValue), "Source of attitude data (AHF/RAF/THF)" )
    endif
    !
    if (hasAttribute(inDataSetH, "ORB_RCNS")) then
       attrValue=stringAttribute(inDataSetH, "ORB_RCNS")
       call setAttribute(PPSOSWDataSetH, "ORB_RCNS", trim(attrValue), "Reconstructed orbit data used?" )
    endif
    !
    if (hasAttribute(inDataSetH, "TFIT_RPD")) then
       attrValue=stringAttribute(inDataSetH, "TFIT_RPD")
       call setAttribute(PPSOSWDataSetH, "TFIT_RPD", trim(attrValue), "Recalculated signal propagation delays?" )
    endif
    !
    if (hasAttribute(inDataSetH, "TFIT_DEG")) then
       int32Value=int32Attribute(inDataSetH, "TFIT_DEG")
       call setAttribute(PPSOSWDataSetH, "TFIR_DEG", int32Value, "Degree of OBT-MET fit polynomial")
    endif
    !
    if (hasAttribute(inDataSetH, "TFIT_RMS")) then
       pa_scx2 = real64Attribute(inDataSetH, "TFIT_RMS")
       call setAttribute(PPSOSWDataSetH, "TFIT_RMS", pa_scx2, "RMS value of OBT_MET polynomial fit" )
    endif
    !
    if (hasAttribute(inDataSetH, "TFIT_PFR")) then
       pa_scx2 = real64Attribute(inDataSetH, "TFIT_PFR")
       call setAttribute(PPSOSWDataSetH, "TFIT_PFR", pa_scx2, "Fraction of disregarded TCS data points" )
    endif
    !
    if (hasAttribute(inDataSetH, "TFIT_IGH")) then
       attrValue=stringAttribute(inDataSetH, "TFIT_IGH")
       call setAttribute(PPSOSWDataSetH, "TFIT_IGH", trim(attrValue), "Ignored GS handover in TC data?" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXPIDSTR")) then
       attrValue=stringAttribute(inDataSetH, "EXPIDSTR")
       call setAttribute(PPSOSWDataSetH, "EXPIDSTR", trim(attrValue), "Exposure identifier" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXPIDSTR")) then
       attrValue=stringAttribute(inDataSetH, "EXPIDSTR")
       call setAttribute(PPSOSWDataSetH, "EXPIDSTR", trim(attrValue), "Exposure identifier" )
    endif
    !
    !
    !
    !
    !*****************************************************************************
    if (hasAttribute(inDataSetH, "DATATYPE")) then
       attrValue=stringAttribute(inDataSetH, "DATATYPE")
       call setAttribute(PPSOSWDataSetH, "DATATYPE", trim(attrValue), "Data type" )
    endif
    !
    if (hasAttribute(inDataSetH, "OBS_ID")) then
       attrValue=stringAttribute(inDataSetH, "OBS_ID")
       call setAttribute(PPSOSWDataSetH, "OBS_ID", trim(attrValue), "Observation Id" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXP_ID")) then
       attrValue=stringAttribute(inDataSetH, "EXP_ID")
       call setAttribute(PPSOSWDataSetH, "EXP_ID", trim(attrValue), "Exposure Id" )
    endif
    !
    if (hasAttribute(inDataSetH, "TMEXP_ID")) then
       int32Value=int32Attribute(inDataSetH, "TMEXP_ID")
       call setAttribute(PPSOSWDataSetH, "TMEXP_ID", int32Value, "Exposure Id from telemetry" )
    endif
    !
    if (hasAttribute(inDataSetH, "SCIWIN")) then
       int32Value=int32Attribute(inDataSetH, "SCIWIN")
       call setAttribute(PPSOSWDataSetH, "SCIWIN", int32Value, "Science window?" )
    endif
    !
    if (hasAttribute(inDataSetH, "PARAM")) then
       int32Value=int32Attribute(inDataSetH, "PARAM")
       call setAttribute(PPSOSWDataSetH, "PARAM", int32Value, "Data Set Parameter" )
    endif
    !
    if (hasAttribute(inDataSetH, "FRMTIME")) then
       int32Value=int32Attribute(inDataSetH, "FRMTIME")
       call setAttribute(PPSOSWDataSetH, "FRMTIME", int32Value, "Frame Time" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATE-OBS")) then
       attrValue=stringAttribute(inDataSetH, "DATE-OBS")
       call setAttribute(PPSOSWDataSetH, "DATE-OBS", trim(attrValue), "Start Time of exposure" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATE-END")) then
       attrValue=stringAttribute(inDataSetH, "DATE-END")
       call setAttribute(PPSOSWDataSetH, "DATE-END", trim(attrValue), "End Time of exposure" )
    endif
    if (hasAttribute(inDataSetH, "BLANK")) then
    !
       int32Value=int32Attribute(inDataSetH, "BLANK")
       call setAttribute(PPSOSWDataSetH, "BLANK", int32Value, "Value indicating missing pixel" )
    endif
    !
    if (hasAttribute(inDataSetH, "BINAX1")) then
       int32Value=int32Attribute(inDataSetH, "BINAX1")
       call setAttribute(PPSOSWDataSetH, "BINAX1", int32Value, "CCD x-axis binning 2**N" )
    endif
    !
    if (hasAttribute(inDataSetH, "BINAX2")) then
       int32Value=int32Attribute(inDataSetH, "BINAX2")
       call setAttribute(PPSOSWDataSetH, "BINAX2", int32Value, "CCD y-axis binning 2**M" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXPOSURE")) then
       int32Value=int32Attribute(inDataSetH, "EXPOSURE")
       call setAttribute(PPSOSWDataSetH, "EXPOSURE", int32Value, "sec", "Exposure duration")
    endif
    !
    if (hasAttribute(inDataSetH, "BINBPE")) then
       attrValue=stringAttribute(inDataSetH, "BINBPE")
       call setAttribute(PPSOSWDataSetH, "BINBPE", trim(attrValue), "BPE binning enabled" )
    endif
    !
    if (hasAttribute(inDataSetH, "FILTER")) then
       attrValue=stringAttribute(inDataSetH, "FILTER")
       call setAttribute(PPSOSWDataSetH, "FILTER", trim(attrValue), "OM filter" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATAMODE")) then
       attrValue=stringAttribute(inDataSetH, "DATAMODE")
       call setAttribute(PPSOSWDataSetH, "DATAMODE", trim(attrValue), "File content" )
    endif

!----------
    call setAttribute(PPSOSWDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
    call setAttribute(PPSOSWDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")

    if (hasAttribute(inDataSetH, "FRAMTIME")) then
       real32Value=real32Attribute(inDataSetH, "FRAMTIME")
       call setAttribute(PPSOSWDataSetH, "FRAMTIME", real32Value, "sec", "Frame time")
    endif

    if (hasAttribute(inDataSetH, "DEADFRAC")) then
       real32Value=real32Attribute(inDataSetH, "DEADFRAC")
       call setAttribute(PPSOSWDataSetH, "DEADFRAC", real32Value, "sec", "Frame dead fraction ")
    endif

!------------------------------------------------------------------------------------------
    outQualityArrayH = addArray(PPSOSWDataSetH, "QUALITY", integer8, dim, "flags", "IMAGE")
    allocate (outQualArray(0 : nxout-1, 0 : nyout-1), stat = allocErr)
    call CheckAllocation(allocErr, "outQualArray")
    outQualArray=0
    !blockH=block(inDataSetH, "MODES", read)
    !call copyBlock(PPSOSWDataSetH, blockH)
!   VNY 11.07.2006 instead of compying the MODES table it should be created
!***************************************************************************
    ! Add a MODES extension, detailing the science window parameters.
    modesTableH = addTable(PPSOSWDataSetH, "MODES", 1)
    columnDim(1) = 1
    modeIMG = 0
!    windowX0a=winX0rot
!    windowY0a=winY0rot
!    windowDXa=winX1rot-winX0rot+1
!    windowDYa=winY1rot-winY0rot+1
!---------------
    windowX0a=winX0ext
    windowY0a=winY0ext
    windowDXa=winX1ext-winX0ext+1
    windowDYa=winY1ext-winY0ext+1

    call AddInt32ColumnG(modesTableH, modeIMG, "MODE", "none", columnDim) 
    call AddInt32ColumnG(modesTableH, windowX0a, "WINDOWX0", "pixels", columnDim) 
    call AddInt32ColumnG(modesTableH, windowY0a, "WINDOWY0", "pixels", columnDim) 
    call AddInt32ColumnG(modesTableH, windowDXa, "WINDOWDX", "pixels", columnDim) 
    call AddInt32ColumnG(modesTableH, windowDYa, "WINDOWDY", "pixels", columnDim) 
    call release(modesTableH)


!***************************************************************************
!----------------------------------------------------------------------------------------

    !********************************************
    ! convert from ra and dec numbers to strings
    !********************************************
 
    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
 
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

    !******************************************************************************
    ! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
    ! origin (IN BINNED PIXELS)
    !******************************************************************************

!07.06.2006 adding the left and bottom margins introduced to the image earlier
    refDX = (xCentre - x0+10) / dXbinning
    refDY = (yCentre - y0+20) / dYbinning
! 02.11.2006 refDX and refDY must me reduced by -(56,630) as measured from the sky
! 07.11.2006 for GRISM2 refDX and refDY must be corrrected by (+146,-723) as measured from the sky

   refDXcorrection=-56. ! this this for GRISM1
   refDYcorrection=-630. ! this is for Grism1
    if (omFilter .eq. "Grism2  " .or. omFilter .eq. "GRISM2  ") then
       refDXcorrection=110. !90. ! 146.
       refDYcorrection=-710. !-723
    endif

    refDX=refDX+refDXcorrection/dXbinning
    refDY=refDY+refDYcorrection/dYbinning

    refDXext=refDX
    refDYext=refDY
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    refx = trans(1) + rotatedRefDX
    refy = trans(4) + rotatedRefDY

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    call setAttribute(PPSOSWDataSetH, "CTYPE1", "PIXELS  ","Coordinate system for X")
    pa_scx2=refx
    call setAttribute(PPSOSWDataSetH, "CRPIX1", pa_scx2, "pix","WCS reference pixel" )
    pa_scx2=ra_point*radiansToDegrees
    call setAttribute(PPSOSWDataSetH, "CRVAL1", pa_scx2, "degrees", "WCS coordinate for the X-ref. pixel" )
    pa_scx2=real(plateScale * dble(xBinning))/3600.0_double
    call setAttribute(PPSOSWDataSetH, "CDELT1", pa_scx2, "degrees", "WCS pixel size" )

    call setAttribute(PPSOSWDataSetH, "CTYPE2", "PIXELS  ","Coordinate system for X")
    pa_scx2=refy
    call setAttribute(PPSOSWDataSetH, "CRPIX2", pa_scx2, "pix","WCS reference pixel" )
    pa_scx2=dec_point*radiansToDegrees
    call setAttribute(PPSOSWDataSetH, "CRVAL2", pa_scx2, "degrees", "WCS coordinate for the Y-ref. pixel" )

    pa_scx2=real(plateScale * dble(yBinning))/3600.0_double
    call setAttribute(PPSOSWDataSetH, "CDELT2", pa_scx2, "degrees", "WCS pixel size" )

    ! 07.09.06 introducing two new keywords to pass information to omatt
    call setAttribute(PPSOSWDataSetH, "OMGXMIN", -trans(1), "pix","X-min of the rotation transformation" )
    call setAttribute(PPSOSWDataSetH, "OMGYMIN", -trans(4), "pix","Y-min of the otation transformation" )
!+++++++++++++++++++++++++++++++++++++++++++++++++


    !-VNY 07.11.2003
    if (produceUndistortedImage) then
!       call SetFitsArrayAttributes(undistArrayH, xBinning, yBinning, plateScale, refx, refy, &
!            ra_scx, dec_scx, ra_point, dec_point)
!
       call SetFitsArrayAttributes(undistArrayH, xBinning, yBinning, plateScale, refDXext, refDYext, &
            ra_scx, dec_scx, ra_point, dec_point)
!
       undistImageP => real32array2data(undistArrayH)
!--------------------------------------------------------------
       extQualityArrayP => int8Array2Data(extQualityArrayH)
       extQualityArrayP=extQualArray
!--------------------------------------------------------------
    else
       allocate (undistImage(0 : nxout-1, 0 : nyout-1), stat = allocErr)
       call CheckAllocation(allocErr, "undistImage")
       undistImage=0
    endif ! produceUndistortedImage


!---
!  : before rotation, the refx and refy should be modified for distortion
    rawx(0) = xCentre       ! position in full OM field.
    rawy(0) = yCentre
!    call CAL_omDistortion(rawx, rawy, deltax, deltay)
! VNY 11.12.2003 using the polinomial approximation 
! with the optional parameter calcMode
    calcMode1=1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)

    !xOrigin=rawx(0)
    !yOrigin=rawy(0)
    !call AddDistortion(xOrigin, yOrigin, xDistorted, yDistorted)
    !refDX=refDX-xDistorted
    !refDY=refDY-yDistorted

!------------------------------- test 14.08.06
    refDX=xCentre/xbinning !192.0_double
    refDY=yCentre/ybinning ! 27.0_double
    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    refx=trans(1) + rotatedRefDX
    refy=trans(4) + rotatedRefDY
!--------------------------------end test 14.08.06
    refDX=refDX-dble(deltax(0))
    refDY=refDY-dble(deltay(0))
!-------------

    ! Compute these offsets in terms of the rotated frame (remember rotation 
    ! takes place about the OM plate centre so the x,y offsets of the OSW origin 
    ! shift in the rotated frame). 

    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    ! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
    ! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
    ! from the BB origin

    refx = trans(1) + rotatedRefDX
    refy = trans(4) + rotatedRefDY
    !
    ! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
    ! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
    !

    ! 26.07.2004 VNY removing the WCS attributes  
    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
    !call SetFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, plateScale, refx, refy, &
    !                              ra_scx, dec_scx, ra_point, dec_point)
    outImage => real32array2data(PPSOSWArrayH)
    !
!------------------------------------------------------
    outQualityArrayP => int8Array2Data(outQualityArrayH)
    outQualityArrayP=outQualArray
!-----------------------------------------------------
    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************
    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)
            k = k + 1
        end do
    end do
    !*********************************************************  
    !  Initialise the undistorted image.  
    !*********************************************************
    if (produceUndistortedImage) then 
       k = 0
       do i = 0, nxdist - 1
          do j = 0, nydist - 1
             call setToNull( undistArrayH, k)
             k = k + 1
          end do
       end do
!       call UndistortImage(nxExt, nyExt, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
!            oswImageExt, undistImageP)
       ! TEST 07.09.06

       testX=197.5_double
       testY=56.5_double
       testX1=itrans(1)+itrans(2)*testX
       testY1=itrans(4)+itrans(5)*testX
       testX1=testX1+itrans(3)*testY
       testY1=testY1+itrans(6)*testY
       
!******* VNY test 18.08.06
!       call UndistortImage4(nxExt, nyExt, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
!            oswImageExt, undistImageP)
!*********
       call UndistortImage3(nxExt, nyExt, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
            oswImageExt, undistImageP)

       call RotateImage(nxdist, nydist, dXbinning, dYbinning, x0, y0, nxout, nyout, &
            undistImageP, outImage, itrans) 
    else
       call UndistortImage3(nxExt, nyExt, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
            oswImageExt, undistImage)
       call RotateImage(nxdist, nydist, dXbinning, dYbinning, x0, y0, nxout, nyout, &
            undistImage, outImage, itrans)
    endif
    call CheckSpectrumExtractability(dXbinning, dYbinning, x0, y0, nxout, nyout, &
            outImage, primInstrValue, omFilter, binBpe, xbin2, ybin2, extractability) 
    call setAttribute(PPSOSWDataSetH, "OMGWIN", extractability, &
         " ", "Spectrum extractability flag (1 is O.K.)" )
    if (hasAttribute(inDataSetH,"DETSEC")) then
       window0 = stringAttribute(inDataSetH, "DETSEC")
       read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
            ch1,winX0, ch2, winX1, ch3,  winY0, ch4, winY1, ch5
       write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",winX0-10/xBinning, ":", winX1+10/xBinning, &
            & ",",winY0-20/yBinning, ":", winY1+20/yBinning, "]" 
       if (xbin .ne. 0. .and. ybin .ne. 0.) then
          write (detsize,'(a6,i4,a6,i4,a1)') "[   1:", INT(2048.0/real(xbin)),",   1:",INT(2048.0/real(ybin)), "]"
       else
          write (detsize,'(a21)') "[   1:2048,   1:2048]"
       endif
       if (produceUndistortedImage)  then
          call setAttribute(undistDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
          call setAttribute(undistDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")
       endif ! produceUndistortedImage
    else
       write(errstr, '(a)') "DETSEC keyword missing"
       call message(AppMsg, Verbose)
    endif

   ! Release handles
    call release(inArrayH)
    if (hasQUALITY) then
       call release(qualArrayH)
    endif
    call release(PPSOSWArrayH)
!---------------------------------- 
    call release(outQualityArrayH)
!---------------------------------
    if (produceUndistortedImage) then
       !--- VNY 11.05.2004
!       call release(extQualityArrayH)
       call release(undistArrayH)
       call release(undistDataSetH)
    endif

    call release(PPSOSWDataSetH)
    call release(inDataSetH)

    !**************************************
    ! Release memory allocation
    !**************************************
    if(allocated(qualArray)) then
       deallocate(qualArray, stat=allocErr)
       call CheckDeallocation(allocErr, "qualArray") 
    endif
    if(allocated(undistImage)) then
       deallocate(undistImage, stat=allocErr)
       call CheckDeallocation(allocErr, "undistImage") 
    endif
    if(allocated(extQualArray)) then
       deallocate(extQualArray, stat=allocErr)
       call CheckDeallocation(allocErr, "extQualArray") 
    endif
    deallocate (oswImageExt, stat=allocErr)
    call CheckDeallocation(allocErr, "oswImageExt")

!if (extractability .lt. 1) then
!   call error("omgprep","The image is too small: the spectrum cannot be extracted")
!endif
    if (extractability .lt. 1) then
       call warning("Window size","The image is too small: the spectrum cannot be extracted")
    endif
end subroutine omgprep


!******************************************
! invert_trans
! inverts the transformation matrix, i.e. gets coefficients to swap between
! x'= t(1) + t(2)*x + t(3)*y
! y'= t(4) + t(5)*x + t(6)*y
! and 
! x= it(1) + it(2)*x' + it(3)*y'
! y= it(4) + it(5)*x' + it(6)*y'
!******************************************
subroutine invert_trans(trans, itrans)
    real(kind=double), intent(in) ::  trans(6)
    real(kind=double), intent(out) :: itrans(6)
    real(kind=double) determ

    determ = trans(2) * trans(6) - trans(3) * trans(5)
    itrans(1) = trans(3) * trans(4) - trans(1) * trans(6)
    itrans(2) = trans(6)
    itrans(3) = -trans(3)
    itrans(4) = trans(1) * trans(5) - trans(2) * trans(4)
    itrans(5) = -trans(5)
    itrans(6) = trans(2)
    itrans = itrans / determ

end subroutine invert_trans
!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
double precision function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(len = 19)  :: dateObs         
    character(len = 19)  :: dateEnd        
    real(kind = double)  :: JulianDate0, JulianDate1, JulianDate2
    character(len = 19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0

end function MissionTime
!*******************************************************************************
! subroutine GetImageBoundaries
! The first step is to find the limits of the sky box that will bound the rotated 
! image, after it is linearised (distortion corrections applied). This is done by 
!
! 1) computing the raw coordinates of the image corners and applying the appropriate 
! distortion correction to them. The coordinates are referenced back to the origin 
! of the undistorted image. The corners are then rotated by the roll angle and 
! converted to binned pixels 
! 2) the limits of this rotated, distorted image are obtained and the translational 
! shift betwen the bounding box origin and the OSW origin are passed to the 
! transformation matrix leading coefficients.
! 3) The dimensions of the bounding box are determined
! 4) The offset of the OM plate centre (1024,1024) are determined relative to 
! the OSW original, as measured in the rotated frame, and then this is 
! corrected to be wrt to the bounding box origin
! 5) WCS keywords are set
! 6) pixels in the sky box are now converted, via the inverse transformation matrix, 
! to the OSW frame and an iterative process to correct back from linearised to raw 
! pixels is used to find the original pixel in the raw image. Bi-linear interpolation 
! is used to fill the rotated image  
! refx and refy are the transformed coordinates of the centre of the image
!*******************************************************************************
subroutine GetImageBoundaries(nx, ny, x0, y0, xBinning, yBinning, &
                          trans, xMin, xMax, yMin, yMax, refx,refy, &
                          filter, primInstrValue, binBpe, xTarget, yTarget)
    integer, intent(in) :: nx, ny   ! The image dimensions
    real(kind=single),  intent(in) :: x0, y0
    integer, intent(in) :: xBinning, yBinning
    real(kind=double), intent(inout), dimension(1 : 6) :: trans   
    real(kind=double), intent(out) :: xMin, xMax, yMin, yMax
    real(kind=single) :: rawx(0 : 0), rawy(0 : 0)      ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)       ! distortion sigmas
    real(kind=double)  :: xlin(4), ylin(4), xcorn(4), ycorn(4)
    integer(kind=int32) :: calcMode1
    real(kind=double), intent(out) :: refx,refy
!    real(kind=double) :: xOrigin, yOrigin, xDistorted, yDistorted

    integer(kind=int32), intent(out) :: xTarget, yTarget
    integer(kind=int32) ::  xbin, ybin
    character(len=6), intent(in) :: filter
    character(len=10), intent(in) :: primInstrValue
    logical(kind=Bool), intent(in) :: binBpe
    xbin=xBinning
    ybin=yBinning

! NOTE definition of CAL_omDistortion means we could set up rawx/y arrays 
! first and then call distortion routine once (passing all corners in at once) and then 
! put loop after to compute lin coords (SRR).

    call GetTargetPosition(primInstrValue,filter,binBpe, xbin, ybin, xTarget, yTarget) 

! distortion of the refpixel
    rawx(0)=xTarget
    rawy(0)=yTarget
    calcMode1=1

    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)

    ! ***** temp 13.08.06 *****************************************************
    xlin(1)=rawx(0)-deltax(0)-x0
    ylin(1)=rawy(0)-deltay(0)-y0
    refx = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)
    refy = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)

    ! ***** temp 13.08.06
    xTarget=rawx(0)-deltax(0)/dble(xBinning)
    yTarget=rawy(0)-deltay(0)/dble(yBinning)

    ! ***** temp 13.08.06 ***************************************************
    rawx(0)=188!xTarget
    rawy(0)=6!yTarget
    calcMode1=1

    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)

    xlin(1)=rawx(0)-deltax(0)-x0
    ylin(1)=rawy(0)-deltay(0)-y0
    refx = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)
    refy = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)

    refx=(xCentre-x0+10) ! unbinned for the time being (at 2048 pixels resolution)
    refy=(xCentre-y0+20) ! 
    rawx(0)=refx
    rawy(0)=refy
    calcMode1=1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
    !**************************************************** NB ***********************
    ! VNY 11.08.06 subtracting twice x0 and y0?
    !**************************************************** NB ***********************
    xlin(1)=rawx(0)-deltax(0)-x0
    ylin(1)=rawy(0)-deltay(0)-y0
    refx = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)
    refy = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)

!========================
! centre pixel
    rawx(0) = x0                              !at 2048 pixels resolution 
    rawy(0) = y0
    calcMode1=1 
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)   
    xlin(1) = rawx(0) - deltax(0) - x0            !correct for distortion and reference to x0,y0 origin
    ylin(1) = rawy(0) - deltay(0) - y0
    !rotate + convert to binned pixels
    xcorn(1) = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)      
    ycorn(1) = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)

!========================
!bottom rh corner in 2048 pixels
    rawx(0) = x0 + real(nx * xBinning)      
    rawy(0) = y0
    calcMode1=1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)   
    xlin(2) = rawx(0) - deltax(0) - x0
    ylin(2) = rawy(0) - deltay(0) - y0

    xcorn(2) = (trans(1) + trans(2) * xlin(2) + trans(3) * ylin(2)) / dble(xBinning)
    ycorn(2) = (trans(4) + trans(5) * xlin(2) + trans(6) * ylin(2)) / dble(yBinning)

!=========================
! top rh corner
    rawx(0) = x0 + real(nx * xBinning)
    rawy(0) = y0 + real(ny * yBinning)
    calcMode1=1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)   
    xlin(3) = rawx(0) - deltax(0) - x0               
    ylin(3) = rawy(0) - deltay(0) - y0

    xcorn(3) = (trans(1) + trans(2) * xlin(3) + trans(3) * ylin(3)) / dble(xBinning)
    ycorn(3) = (trans(4) + trans(5) * xlin(3) + trans(6) * ylin(3)) / dble(yBinning)

!======================
! top lh corner
    rawx(0) = x0
    rawy(0) = y0 + real(ny * yBinning)
    calcMode1=1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)   
    xlin(4) = rawx(0) - deltax(0) - x0
    ylin(4) = rawy(0) - deltay(0) - y0

    xcorn(4) = (trans(1) + trans(2) * xlin(4) + trans(3) * ylin(4)) / dble(xBinning)
    ycorn(4) = (trans(4) + trans(5) * xlin(4) + trans(6) * ylin(4)) / dble(yBinning)


!  We now have the corners of OSW in linearised coordinates relative to the origin 
!  of the original OSW image, after rotation of the frame 
!  Now need to get limits of bounding box (BB)  from these corners (in rotated frame)
!  Once minimum x and y are found (origin of BB - i.e. the rotated frame), since these 
!  are wrt the original OSW origin, we can determine the offset of the OSW origin from 
!  bounding BB (transfer offsets to trans(1), trans(4))

    xmax = trans(1)
    xmin = trans(1)
    call GetMinMax(xmin, xmax, xcorn)
    trans(1) = trans(1) - xmin                        !get x offset of OSW origin from BB origin

    ymax = trans(4)
    ymin = trans(4)
    call GetMinMax(ymin, ymax, ycorn)
    
    trans(4) = trans(4) - ymin                        !get y offset of OSW origin from BB origin
 
    refx=refx+trans(1)
    refy=refy+trans(4)

end subroutine GetImageBoundaries
!******************************
! subrioutine GetMinMax
!******************************
subroutine GetMinMax(minimum, maximum, corner)
    real(kind=double), intent(inout) :: minimum
    real(kind=double), intent(inout) :: maximum
    real(kind=double), intent(in), dimension(1 : 4) :: corner
    integer :: i

    do i = 1, 4
        minimum = dmin1(minimum, corner(i))
        maximum = dmax1(maximum, corner(i))
    end do

end subroutine GetMinMax 
!*****************************************
! subroutine SetFitsArrayAttributes
!*****************************************
subroutine SetFitsArrayAttributes(arrayH, xBinning, yBinning,     &
                                  plateScale, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point)

    type(ArrayT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(kind=double), intent(in) :: plateScale
    real(kind=single), intent(in) :: refx, refy
    real(kind=double), intent(in) :: ra_scx, dec_scx
    real(kind=double), intent(in) :: ra_point, dec_point


    call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
    call setAttribute(arrayH, "CRPIX1", refx, "WCS reference pixel") 
    call setAttribute(arrayH, "RA_SCX", ra_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for X")
    call setAttribute(arrayH, "CRVAL1", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "CDELT1", (-1.0)*plateScale*(real(xBinning) * radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
    call setAttribute(arrayH, "CRPIX2", refy, "WCS reference pixel")
    call setAttribute(arrayH, "DEC_SCX", dec_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for Y")
    call setAttribute(arrayH, "CRVAL2", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
    call setAttribute(arrayH, "CDELT2", plateScale*(real(yBinning)*radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetFitsArrayAttributes


!***************************************************************
! Subroutine RotateImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
subroutine  RotateImage(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, rotatedImage, &
                      itrans)  
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: rotatedImage
    real(kind=double), intent(in)  :: itrans(6)     ! its inverse
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: t, u, rx0, ry0, rx, ry
    real               :: rxInit, ryInit
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(kind=double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)       ! distortion sigmas
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: radius1, radius2, oldDiff, newDiff
    real(kind=double) :: rxOld, ryOld
    integer :: i, j, j0, k0 
    integer :: j1, k1, iter, istat
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    integer(kind=int32) :: calcMode1
!    real(kind=double) :: xOrigin, yOrigin, xDistorted, yDistorted

!setting a value to all pixels of the rotated image
    rotatedImage=-1.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")
    
    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 
        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + 0.5_double)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + 0.5_double)
    end forall


    dblNx = dble(nx)
    dblNy = dble(ny)

    do i = 0, nxout - 1
       rx0 = itrans(1) + itrans(2) * (dble(i) + 0.5_double)
       ry0 = itrans(4) + itrans(5) * (dble(i) + 0.5_double)
     
       do j = 0, nyout - 1

          !***************************************************************      
          !  use inverse transformation to get coords in unrotated 
          ! (input) frame note part of this is already done within i loop
          !***************************************************************
          rx = rx0 + itrans3TimesjPlusPoint5(j)
          ry = ry0 + itrans6TimesjPlusPoint5(j)

          !****************Test 070906 *****************
          !***************************************************************************
          !  Calculate position in full frame. Get distortion correction and correct 
          !  positions. THIS is done iteratively (twice) because we want to go from 
          !  linearised to raw position but CAL distortion correction routine gives 
          !  values based on raw input positions. Thus we correct linearised position 
          !  with wrong correction values. To improve the final position we apply 
          !  the wrong distortion correction to get new coordinates and then recompute 
          !  the corrections using this position. This should bring us close to the 
          !  true raw position. 
          !
          ! VNY 10.09.2001 correction for the half of the pixel as the CCD 
          ! grid passes through the pixel center
          !****************************************************************** 
        
          rxInit = rx
          ryInit = ry
          oldDiff = 1e6

          do iter = 1, maxIterations
             rawx(0) = (rx - 0.5) * dXbinning + x0       ! position in full OM field.
             rawy(0) = (ry - 0.5) * dYbinning + y0


!             call CAL_omDistortion(rawx, rawy, deltax, deltay)
             calcMode1=1
             call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
!             xOrigin=rawx(0)
!             yOrigin=rawy(0)
!             call AddDistortion(xOrigin, yOrigin, xDistorted, yDistorted)
!             rx=rxInit+xDistorted/dXbinning
!             ry=ryInit+yDistorted/dYbinning
             rx = rxInit + deltax(0) / dXbinning   ! 25.8.2000 The signs are +ve because we are going from linear 
             ry = ryInit + deltay(0) / dYbinning   ! to raw positions. Convert back to binned pixels.

             ! VNY 19.11.02 The two above lines result in a shift in x-direction
             ! somwhere in the middle of the image (probably distortion is wrong)

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& VNY 19.11.2002
             rx=rxInit
             ry=ryInit
!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

             radius1 = sqrt(rx * rx + ry * ry)
             radius2 = sqrt(rxInit * rxInit + ryInit * ryInit)
             newDiff = abs(radius1 - radius2)


             if (newDiff .gt. oldDiff) then
                 rx = rxOld
                 ry = ryOld

                 exit
             else  
                 if (newDiff  .lt. 0.05) then
                 
                     exit
                 
                 endif 
             end if
             oldDiff = newDiff
             rxOld = rx
             ryOld = ry
          end do

          !*********************************************
          ! use bi-linear interpolation to fill in grid
          !*********************************************
          if (rx .ge. 0.0_double .and. rx .le. dblNx .and. & 
              ry .ge. 0.0_double .and. ry .le. dblNy) then

              !********************************************************************
              !   VNY 10.09.2001 correction for the half of the pixel because
              !   the origin of the CCD coordinate frame is in the pixel centre 
              !   
              !********************************************************************
              rx = rx - 0.5_double
              ry = ry - 0.5_double

              j0 = max(0, int(rx))
              j0 = min(nxMinus2, j0)
              k0 = min(nyMinus2, int(ry))
              k0 = max(0, k0)
              j1 = j0 + 1
              k1 = k0 + 1

              t = (rx - dble(j0))  
              u = (ry - dble(k0))

              rotatedImage(i, j ) = (1.0_double - t) * (1.0_double - u) * oswImage(j0, k0) + &
                                    t                * (1.0_double - u) * oswImage(j1, k0) + &
                                    t                * u                * oswImage(j1, k1) + &
                                   (1.0_double - t)  * u                * oswImage(j0, k1)

          end if
       end do
    end do

    deallocate(itrans3TimesjPlusPoint5, stat = istat)
    call  CheckDeAllocation(istat, "itrans3TimesjPlusPoint5")
 
    deallocate(itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans6TimesjPlusPoint5")
 

end subroutine RotateImage

!***************************************************************
! Subroutine CheckSpectrumExtractability
!
! This subroutine will check the extractability of the default 
! (target) spectrum. This is needed by the grism-chain in order
! to decide whether to run the detection and extraction task on
! this image or not (the image could be too small, which would 
! preclude the extraction of a useful spectrum) 
!****************************************************************
subroutine  CheckSpectrumExtractability(dXbinning, dYbinning, x0, y0, & 
                      nx, ny, rotatedImage, primInstrValue, filter, &
                      binBpe, xbin, ybin, extractability)  
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: rotatedImage
    character(len=10), intent(in) :: primInstrValue
    integer, intent(out) :: extractability
    character(len=6), intent(in) :: filter
    logical(kind=Bool), intent(in) :: binBpe
    integer(kind=int32), intent(in) :: xbin,ybin
    integer(kind=int32) :: xTarget, yTarget, i1
    real(kind=double) :: yZeroref
    real(kind=double) :: yFirst
    real(kind=double) :: source2spectrum, spectrumLength
    integer :: iyUp, iyDown, iyUp2
    real(kind=double) :: src2spectrumU, src2spectrumV
    real(kind=double) :: spectrumToleranceU, spectrumToleranceV, spectrumTolerance
    real(kind=double) :: spectrumLengthParU, spectrumLengthParV
    real(kind=double) :: availableSpectrumFraction, availableSpectrumLength

    src2spectrumU = 900.0 
    src2spectrumV = 785.0
    spectrumToleranceU= 600.0
    spectrumToleranceV=400.0
    spectrumLengthParU = 1100.0
    spectrumLengthParV = 600.0

    ! Checking the specrum regions for both filters
    if (trim(filter) .eq. "Grism2" .or. trim(filter) .eq. "GRISM2")  then ! (visual grism)
       source2spectrum=src2spectrumV ! 800.0
       spectrumLength=spectrumLengthParV !700.0
       spectrumTolerance=spectrumToleranceV
       if (ybin .gt. 1.0) then 
          source2spectrum=source2spectrum/ybin
          spectrumLength=spectrumLength/ybin
          spectrumTolerance=spectrumTolerance/ybin
       endif
    endif
    
    if (trim(filter) .eq. "Grism1" .or. trim(filter) .eq. "GRISM1")  then ! (UV grism)
       source2spectrum=src2spectrumU ! 800.0 
       spectrumLength=spectrumLengthParU ! 700.0
       spectrumTolerance=spectrumToleranceU
       if (ybin .gt. 1.0) then 
          source2spectrum=source2spectrum/ybin
          spectrumLength=spectrumLength/ybin
          spectrumTolerance=spectrumTolerance/ybin
       endif
    endif

    extractability=1

    ! Getting the target position
    call GetTargetPosition(primInstrValue,filter,binBpe, xbin, ybin, xTarget, yTarget) 
    ! VNY 02.08.2006 xTarget and yTarget are already given taking into account
    ! the binning of the image 
    !i1=xTarget/ybin
    i1=xTarget
    if (i1 .gt. nx-1) i1=nx-1
    !yZeroRef=yTarget/ybin
    yZeroRef=yTarget

    yFirst=yTarget+source2spectrum
    iyUp=int(yFirst)
    iyDown=int(yFirst-spectrumLength/2.0)

    if (iyUp .ge. ny-2) iyUp=ny-2
    if (iyDown .ge. iyUp) iyDown=iyUp-1

    availableSpectrumLength=0.0
    availableSpectrumFraction=0.0

    if (iyDown .gt. yTarget .and. iyDown .lt. ny-1 .and. iyUp .lt. ny-1) then
       if (rotatedImage(i1,iyDown) .gt. -0.9 .and. rotatedImage(i1,iyUp) .gt. -0.9) then
          ! determine the upper available limit
          availableSpectrumLength=0.0
          checkUpA: do iyUp2=iyDown+1,iyDown+int(spectrumLength)
             if (iyUp2 .lt. ny) then
                if (rotatedImage(i1,iyUp2) .gt. -0.9) then
                   availableSpectrumLength=availableSpectrumLength+1.0
                else
                   exit checkUpA
                endif
             endif
          enddo checkUpA
       endif               
    endif
    if (spectrumLength .gt. 0) then
       availableSpectrumFraction=availableSpectrumLength/spectrumLength
    else
       availableSpectrumFraction=0
    endif
    ! Extract the spectrum only if more than 50 percent of the spectrum is available
    if (availableSpectrumFraction .le. 0.5) extractability=0

end subroutine CheckSpectrumExtractability

!****************************************************************
! Subroutine to get the X and Y coordinates of the target
!****************************************************************
subroutine GetTargetPosition(primInstrValue, filter, binBpe, xbin, ybin, xTarget, yTarget)
    character(len=*), intent(in) :: primInstrValue, filter
    logical(kind=Bool), intent(in) :: binBpe
    integer(kind=int32), intent(in) :: xbin, ybin
    integer(kind=int32), intent(out) :: xTarget, yTarget
    integer(kind=int32), dimension(3,2,2) :: xCal, yCal
    integer :: i,j,k

! The calibration data for the target measured by A.Talavera 20.07.2004
    data xCal/236,252,211,242,209,205,616,618,560,602,603,545/
    data yCal/ 67, 65,124, 51,102,102,232,251,753,196,205,239/
!-------------------------------------------------------------
!  Only for test purporses 27.07.2004:
!    data xCal/236,252,211,242,209,205,616,618,560,602,150,545/
!    data yCal/ 67, 65,124, 51,102,102,232,251,753,196, 80,239/
!
!  Current table for 20.07.2004
!-------------------------------------------------------------------
!        R1            PN                OM       <-- Prim.instr
!-------------------------------------------------------------------
!     X     Y       X       Y         X      Y
!-------------------------------------------------------------------
!   236    67      252      65       211    124   <-- UV, HR
!   242    51      209     102       205    102   <--  V, HR
!
!   616   232      618     251       560    753   <-- UV, LR
!   602   196      603     205       545    239   <--  V, LR
!-------------------------------------------------------------------     
    i=2 ! primary instrument: PN or E3
    if (trim(primInstrValue) .eq. "R1") i=1
    if (trim(primInstrValue) .eq. "OM") i=3
    
    j=1 ! filter=UV (grism1)
    if (trim(filter) .eq. "GRISM2" .or. trim(filter) .eq. "Grism2") j=2

    k=1 ! resolution is High (binBpe=F)
    if (binBpe) k=2
    xTarget=xCal(i,j,k)
    yTarget=yCal(i,j,k)
    if (xbin .gt. 1 .and. .not. binBpe) xTarget=xTarget/xbin
    if (ybin .gt. 1 .and. .not. binBpe) yTarget=yTarget/ybin
end subroutine GetTargetPosition

!***************************************************************
! Subroutine GetPrimaryInstrument
!
! This subroutine will find the primary instrument name using
! the summary-file
!****************************************************************
subroutine  GetPrimaryInstrument(PPSOSWSet,primInstrValue) 
  character(len=255), intent(in) :: PPSOSWSet   
  character(len=10), intent(out) :: primInstrValue
  integer :: nExps, count1
  character(len=1024)  :: summaryFileName, summaryShortFileName, outDirectory
  character(len=1024) :: modifiedSummaryFileName, summaryString
  integer(kind=int16) :: posPoint, primPos, pos1, pos2
  logical :: existFile, eofReached, primInstrFound, existModifiedFile
  logical :: existWithinOutDirectory
       
  nExps = OAL_selectScope(XMM, SUMMARY_INFORMATION)   

  !primeInstrString  = OAL_getIPPVstring("PRIME_INSTR")
  call OAL_selectFile(0, summaryFileName);
  posPoint=0
  posPoint=index(trim(summaryFileName),'/',.true.) ! .true. means backward
  summaryShortFileName=summaryFileName 
  if (posPoint .gt. 3) then
     summaryShortFileName(1:posPoint)=' '
     summaryShortFileName=adjustl(summaryShortFileName)
  endif

  ! Find the name of the output directory
  posPoint=0        
  outDirectory=PPSOSWSet
  posPoint=index(trim(outDirectory),'/',.true.) ! .true. means backward
  if (posPoint .ge. 1) then
     outDirectory(posPoint:1024)=' '
     ! combine the name of the summary file and the output directory
     outDirectory(posPoint:posPoint+1)='/ '
     outDirectory(posPoint+1:1024)=summaryShortFileName(1:1024-posPoint)
  else
     outDirectory='./'
     outDirectory(3:1024)=summaryShortFileName(1:1021)
  end if
  existFile=.false.
  existModifiedFile=.false.
  existWithinOutDirectory=.false.
  primPos=0
  primPos=index(trim(summaryFileName),'SUM.')
  ! Find summary file in the usual directory (ODF)
  inquire (FILE=trim(summaryFileName), EXIST=existFile)
  if (.not. existFile) then
     ! Search for the original (ascii) summary file 
     modifiedSummaryFileName=trim(summaryFileName)
     if (modifiedSummaryFileName(primPos+4:primPos+6) .eq. 'SAS') then
        modifiedSummaryFileName(primPos+4:primPos+6)='ASC'
     else
        modifiedSummaryFileName(primPos+4:primPos+6)='SAS'
     endif
     inquire(FILE=trim(modifiedSummaryFileName), EXIST=existModifiedFile)
     if (.not. existModifiedFile) then
        inquire(FILE=trim(outDirectory),EXIST=existWithinOutDirectory)
     endif
  endif

  if (existFile .or. existModifiedFile .or. existWithinOutDirectory) then
     ! Open the summary file
     if (existFile) open(2,file=trim(summaryFileName))
     if (existModifiedFile) open(2,file=trim(modifiedSummaryFileName))
     if (existWithinOutDirectory) open(2,file=trim(outDirectory))
     eofReached=.false.
     primInstrFound=.false.
     count1=0
     primPos=0
     primInstrValue=''
     do
        read(2, '(a)', end=110) summaryString
        count1=count1+1
        primPos=index(summaryString,'PRIME_INSTR')
        if (primPos .gt. 0) then
           primInstrFound=.true.
           exit
        endif
     enddo
110  continue
120  close(2)         
     if (primInstrFound) then
        pos1=index(summaryString, '=')
        pos2=index(summaryString, '/')
        primInstrValue=summaryString(pos1+2:pos2-1)
     else
        primInstrValue=''
        call warning("SummaryFile", "Primary instrument string not found in the summary file")
     endif
  else
     primInstrValue=''
     call warning("SummaryFile", "No summary file found")
  endif

end subroutine GetPrimaryInstrument

!***************************************************************
! Subroutine UndistortImage
!
!****************************************************************
subroutine  UndistortImage(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, undistortedImage)
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: undistortedImage
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: rx, ry!, rx0, ry0, t, u 
    real               :: rxNew, ryNew, di, dj
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)     
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: dblNxOut, dblNyOut
    integer :: i, j
    integer :: iout, jout
    integer :: i1, j1
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    real(kind=double) :: totImage, leftImage, rightImage, leftUpImage, leftDownImage, rightUpImage, rightDownImage
    integer(kind=int32) :: calcMode1

!setting a value to all pixels of the rotated image

    undistortedImage=0.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    
    dblNx = dble(nx)
    dblNy = dble(ny)
! VNY 22.07.03
    dblNxOut = dble(nxout)
    dblNyOut = dble(nyout)
    do i = 0, nx - 1
       rx = dble(i) !+ 0.5_double
       do j = 0, ny - 1
          ry = dble(j) !+ 0.5_double 
          rawx(0) = (rx + x0)*dXbinning       ! position in full OM field.
          rawy(0) = (ry + y0)*dYbinning
          calcMode1=1
          call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
          rxNew = rx - deltax(0) /dXbinning     ! 25.8.2000 The signs are +ve because we are going from linear 
          ryNew = ry - deltay(0) /dYbinning   ! to raw positions. Convert back to binned pixels.

          iout=int(rxNew)
          jout=int(ryNew)
          di=rxNew-real(iout)
          dj=ryNew-real(jout)
          i1=1
          if (di .lt. 0) then
             i1=-1
             di = -di
          endif
          j1=1
          if (dj .lt. 0) then
             j1=-1
             dj=-dj
          endif


          ! Subdivision of the image in four parts
          totImage=oswImage(i,j)
          leftImage=totImage*di+0.5
          rightImage=totImage-leftImage
          leftDownImage=leftImage*dj+0.5
          leftUpImage=leftImage-leftDownImage
          rightDownImage=rightImage*dj+0.5
          rightUpImage=rightImage-rightDownImage
          

          if (iout .lt. nxout .and. iout .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
             undistortedImage(iout,jout)= undistortedImage(iout,jout)+rightUpImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
             undistortedImage(iout+i1,jout)= undistortedImage(iout+i1,jout)+leftUpImage
          endif
          if (iout .lt. nxout .and. iout .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
             undistortedImage(iout,jout+j1)= undistortedImage(iout,jout+j1)+rightDownImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
             undistortedImage(iout+i1,jout+j1)= undistortedImage(iout+i1,jout+j1)+leftDownImage
          endif
       end do
    end do
end subroutine UndistortImage


!***************************************************************
! Subroutine UndistortImage2
! Undistortion by sub-dividing the pixels into small fractions 
!
!****************************************************************
subroutine  UndistortImage2(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, undistortedImage)
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: undistortedImage
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: rx, ry!, rx0, ry0, t, u 
    real               :: rxNew, ryNew, di, dj
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)     
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: dblNxOut, dblNyOut
!    real(kind=double) :: radius2, oldDiff, newDiff, radius1
!    real(kind=double) :: rxOld, ryOld
    integer :: i, j!, k0, j0
    integer :: iout, jout
    integer :: i1, j1!, istat, iter, k1
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    real(kind=double) :: totImage, leftImage, rightImage, leftUpImage, leftDownImage, rightUpImage, rightDownImage
    integer(kind=int32) :: calcMode1
!    real(kind=double) :: xOrigin, yOrigin, xDistorted, yDistorted
!---VNY Test 31.05.2005

!    real(kind=single), dimension(:, :), allocatable :: tmpArray
    integer :: nxTmp, nyTmp
    integer(kind=int32) :: allocErr
    real(kind=single) :: value10
    integer :: i2,j2
    real :: rnum

    nxTmp=nxout*30 !###
    nyTmp=nyOut*30 !###

!setting a value to all pixels of the rotated image
    undistortedImage=0.0


    do i1=0,nxTmp-1
       i=i1/30 !###
       rx=dble(i1)/30.0_double !###
       do j1=0,nyTmp-1
          ! Compute the pixel position of the input image

          j=j1/30 !###
          value10=oswImage(i,j)/900.0 !###
          ry=dble(j1)/30.0_double !###
          rawx(0)=(rx+x0)*dXbinning
          rawy(0)=(ry+y0)*dYbinning
          calcMode1=1
          call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
          rxNew=rx-deltax(0)/dXbinning+0.5
          ryNew=ry-deltay(0)/dYbinning+0.5
          call random_number(rnum)
!          if (rnum .gt. 0.5) then
             rxNew=rxNew+0.1*(0.5-rnum)
!          else
!             rxNew=rxNew-1.0
!          endif
          call random_number(rnum)
!          if (rnum .gt. 0.5) then
             ryNew=ryNew+0.1*(0.5-rnum)
!          else
!             ryNew=ryNew-1.0
!          endif
          !rxNew=rxNew*10.0
          !ryNew=ryNew*10.0
          i2=int(rxNew)
          j2=int(ryNew)
          if (i2 .lt. nxout .and. i2 .ge. 0 .and. j2 .lt. nyout .and. j2 .ge. 0)  then
             undistortedImage(i2,j2)= undistortedImage(i2,j2)+value10
          endif
       enddo
    enddo
end subroutine UndistortImage2

!***************************************************************
! Subroutine UndistortImage3
!
!****************************************************************
subroutine  UndistortImage3(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, undistortedImage)
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: undistortedImage
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: rx, ry!, rx0, ry0, t, u 
    real               :: rxNew, ryNew, di, dj
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)     
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: dblNxOut, dblNyOut
    integer :: i, j
    integer :: iout, jout, iout0, jout0
    integer :: i1, j1
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    real(kind=single) :: totImage, leftImage, rightImage, leftUpImage, leftDownImage, rightUpImage, rightDownImage
    integer(kind=int32) :: calcMode1

    integer, dimension(:, :), allocatable :: rda!,lda,lua,rua
    integer(kind=int32) :: allocErr
    integer value1, njumpsI, njumpsJ, njumpsIJ
    logical(kind=bool) :: jumpi, jumpj
    real(kind=single) :: rvalue1
    real :: rnum

 !   allocate (lua(0 : nxout-1, 0 : nyout-1), stat = allocErr)
 !   call CheckAllocation(allocErr, "lua array")
 !   allocate (lda(0 : nxout-1, 0 : nyout-1), stat = allocErr)
 !   call CheckAllocation(allocErr, "lda array")
 !   allocate (rua(0 : nxout-1, 0 : nyout-1), stat = allocErr)
 !   call CheckAllocation(allocErr, "lua array")
    allocate (rda(0 : nxout-1, 0 : nyout-1), stat = allocErr)
    call CheckAllocation(allocErr, "lua array")

 !   lua=0
 !   lda=0
 !   rua=0
    rda=0

    
!setting a value to all pixels of the rotated image
    undistortedImage=0.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    
    dblNx = dble(nx)
    dblNy = dble(ny)
! VNY 22.07.03
    dblNxOut = dble(nxout)
    dblNyOut = dble(nyout)
    iout0=0
    jout0=0
    njumpsI=0
    njumpsJ=0
    njumpsIJ=0
    do i = 0, nx - 1
       rx = dble(i) !+ 0.5_double
       do j = 0, ny - 1
          ry = dble(j) !+ 0.5_double 
          rawx(0) = (rx + x0)*dXbinning       ! position in full OM field.
          rawy(0) = (ry + y0)*dYbinning
          calcMode1=1
          call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
          rxNew = rx - deltax(0) /dXbinning     ! 25.8.2000 The signs are +ve because we are going from linear 
          ryNew = ry - deltay(0) /dYbinning   ! to raw positions. Convert back to binned pixels.


!-----------------------
          iout=floor(rxNew)
          jout=floor(ryNew)
!----------------
          jumpi=.false.
          jumpj=.false.
          if (iout .gt. iout0+1) then
             njumpsI=njumpsI+1
             jumpi=.true.
          endif
          if (jout .gt. jout0+1) then
             njumpsJ=njumpsJ+1
             jumpj=.true.
          endif
          iout0=iout
          jout0=jout
!----------------
          
          di=rxNew-real(iout)
          dj=ryNew-real(jout)
!------------- 1.06.2005
!          if (dj .lt. 0.1 .or. dj .gt. 0.9) then
!             call random_number(rnum)
!             ryNew=ryNew+rnum*0.1
!             jout=floor(ryNew)
!             dj=ryNew-real(jout)
!          endif
!
!          if (dj .gt. 0.9) then
!             dj=0.999
!          endif
!-------------
          i1=1
          if (di .lt. 0) then            
                i1=-1
                di = -di
          endif
          j1=1
          if (dj .lt. 0) then
                j1=-1
                dj=-dj
          endif

          ! Subdivision of the image in four parts (bi-linear interpolation)
          totImage=oswImage(i,j)
          leftImage=totImage*di
          rightImage=totImage-leftImage
          leftDownImage=leftImage*dj
          leftUpImage=leftImage-leftDownImage
          rightDownImage=rightImage*dj
          rightUpImage=rightImage-rightDownImage



         if (iout .lt. nxout .and. iout .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                undistortedImage(iout,jout)= undistortedImage(iout,jout)+rightUpImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                undistortedImage(iout+i1,jout)= undistortedImage(iout+i1,jout)+leftUpImage
          endif
          if (iout .lt. nxout .and. iout .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
                undistortedImage(iout,jout+j1)= undistortedImage(iout,jout+j1)+rightDownImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
                undistortedImage(iout+i1,jout+j1)= undistortedImage(iout+i1,jout+j1)+leftDownImage
          endif   
!-----------     
     


          ! label the overjumped pixels
          !if (jumpi) then
          !   if (iout .lt. nxout .and. iout.ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
          !      lda(iout,jout)=1
          !   endif
          !endif
          if (jumpj) then
             if (iout .lt. nxout .and. iout.ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                rda(iout,jout)=1
             endif
          endif

       end do
    end do


    ! VNY 01.06.2005 remove the pattern that appear due to the moire effect
    do i=0,nxout-1
       do j=3,nyout-3
          if (rda(i,j) .eq. 1) then
             rvalue1=undistortedImage(i,j-3)+undistortedImage(i,j-2)
             rvalue1=rvalue1+undistortedImage(i,j)+undistortedImage(i,j+1)
             rvalue1=rvalue1/4.0
             undistortedImage(i,j-1)=rvalue1
             rvalue1=undistortedImage(i,j-2)+undistortedImage(i,j-1)
             rvalue1=rvalue1+undistortedImage(i,j+1)+undistortedImage(i,j+2)
             rvalue1=rvalue1/4.0
             undistortedImage(i,j)=rvalue1
          endif
       enddo
    enddo

        

    !if(allocated(lua)) then
    !   deallocate(lua, stat=allocErr)
    !   call CheckDeallocation(allocErr, "lua") 
    !endif
    !if(allocated(lda)) then
    !   deallocate(lda, stat=allocErr)
    !   call CheckDeallocation(allocErr, "lda") 
    !endif
    !if(allocated(rua)) then
    !   deallocate(rua, stat=allocErr)
    !   call CheckDeallocation(allocErr, "rua") 
    !endif
    if(allocated(rda)) then
       deallocate(rda, stat=allocErr)
       call CheckDeallocation(allocErr, "rda") 
    endif

end subroutine UndistortImage3

!***************************************************************
! Subroutine UndistortImage4
!
!****************************************************************
subroutine  UndistortImage4(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, undistortedImage)
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: undistortedImage
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: rx, ry!, rx0, ry0, t, u 
    real               :: rxNew, ryNew, di, dj
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=single), pointer     :: sigmax(:), sigmay(:)     
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: dblNxOut, dblNyOut
    integer :: i, j
    integer :: iout, jout, iout0, jout0
    integer :: i1, j1
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    real(kind=single) :: totImage, leftImage, rightImage, leftUpImage, leftDownImage, rightUpImage, rightDownImage
    integer(kind=int32) :: calcMode1

    integer, dimension(:, :), allocatable :: rda!,lda,lua,rua
    integer(kind=int32) :: allocErr
    integer value1, njumpsI, njumpsJ, njumpsIJ
    logical(kind=bool) :: jumpi, jumpj
    real(kind=single) :: rvalue1
    real :: rnum


    allocate (rda(0 : nxout-1, 0 : nyout-1), stat = allocErr)
    call CheckAllocation(allocErr, "lua array")
    rda=0
    
!setting a value to all pixels of the undistorted image
    undistortedImage=0.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3
    
    dblNx = dble(nx)
    dblNy = dble(ny)
! VNY 22.07.03
    dblNxOut = dble(nxout)
    dblNyOut = dble(nyout)
    iout0=0
    jout0=0
    njumpsI=0
    njumpsJ=0
    njumpsIJ=0
    do i = 0, nx - 1
       rx = dble(i) !+ 0.5_double
       do j = 0, ny - 1
          ry = dble(j) !+ 0.5_double 
          rawx(0) = (rx + x0)*dXbinning       ! position in full OM field.
          rawy(0) = (ry + y0)*dYbinning
          calcMode1=1
          call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, calcMode1)
          rxNew = rx! - deltax(0) /dXbinning     ! 25.8.2000 The signs are +ve because we are going from linear 
          ryNew = ry! - deltay(0) /dYbinning   ! to raw positions. Convert back to binned pixels.

!-----------------------
          iout=floor(rxNew)
          jout=floor(ryNew)
!----------------
          jumpi=.false.
          jumpj=.false.
          if (iout .gt. iout0+1) then
             njumpsI=njumpsI+1
             jumpi=.true.
          endif
          if (jout .gt. jout0+1) then
             njumpsJ=njumpsJ+1
             jumpj=.true.
          endif
          iout0=iout
          jout0=jout
!----------------
          
          di=rxNew-real(iout)
          dj=ryNew-real(jout)

          i1=1
          if (di .lt. 0) then            
                i1=-1
                di = -di
          endif
          j1=1
          if (dj .lt. 0) then
                j1=-1
                dj=-dj
          endif

          ! Subdivision of the image in four parts (bi-linear interpolation)
          totImage=oswImage(i,j)
          leftImage=totImage*di
          rightImage=totImage-leftImage
          leftDownImage=leftImage*dj
          leftUpImage=leftImage-leftDownImage
          rightDownImage=rightImage*dj
          rightUpImage=rightImage-rightDownImage

         if (iout .lt. nxout .and. iout .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                undistortedImage(iout,jout)= undistortedImage(iout,jout)+rightUpImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                undistortedImage(iout+i1,jout)= undistortedImage(iout+i1,jout)+leftUpImage
          endif
          if (iout .lt. nxout .and. iout .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
                undistortedImage(iout,jout+j1)= undistortedImage(iout,jout+j1)+rightDownImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
                undistortedImage(iout+i1,jout+j1)= undistortedImage(iout+i1,jout+j1)+leftDownImage
          endif   
!-----------     
          if (jumpj) then
             if (iout .lt. nxout .and. iout.ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
                rda(iout,jout)=1
             endif
          endif

       end do
    end do


    ! VNY 01.06.2005 remove the pattern that appear due to the moire effect
    do i=0,nxout-1
       do j=3,nyout-3
          if (rda(i,j) .eq. 1) then
             rvalue1=undistortedImage(i,j-3)+undistortedImage(i,j-2)
             rvalue1=rvalue1+undistortedImage(i,j)+undistortedImage(i,j+1)
             rvalue1=rvalue1/4.0
             undistortedImage(i,j-1)=rvalue1
             rvalue1=undistortedImage(i,j-2)+undistortedImage(i,j-1)
             rvalue1=rvalue1+undistortedImage(i,j+1)+undistortedImage(i,j+2)
             rvalue1=rvalue1/4.0
             undistortedImage(i,j)=rvalue1
          endif
       enddo
    enddo

    if(allocated(rda)) then
       deallocate(rda, stat=allocErr)
       call CheckDeallocation(allocErr, "rda") 
    endif

end subroutine UndistortImage4


!********************************************
! Subroutine ModifySourcePositionAngles
!*******************************************
subroutine ModifySourcePositionAngles(tableSetH, angle, nullValue)
    type(TableT), intent(in) :: tableSetH
    real(kind=double), intent(in) :: angle
    real(kind=single), intent(in) :: nullValue
    
    real(kind=single), pointer :: positionAngle(:)
    integer(kind=int32) :: noRows, position
  
    noRows = numberOfRows(tableSetH)

    position = columnNumber(tableSetH, "PA")

    positionAngle => AllocateSingle(1, noRows, "positionAngle")
    call ReadReal32Column(tableSetH, "PA", noRows, 0, positionAngle)
 
    call deleteColumn(tableSetH, "PA")

    positionAngle = positionAngle + angle
    call AddReal32Column(tableSetH, positionAngle, noRows, "PA", "arcsecs", &
             "PA of major-axis Referred to equatorial system", nullvalue, position)
 

    deallocate(positionAngle)
end subroutine ModifySourcePositionAngles


!********************************************************
! subroutine smoothIMage
! Routine to smooth image 
!
!********************************************************
!subroutine SmoothImage1(image, nx, ny, XsmoothSize, YsmoothSize)
subroutine SmoothImage(image, nx, ny, XsmoothSize, YsmoothSize)
 
    integer (int32), intent(in)    :: nx, ny
    real (single), pointer      :: image(:, :)
    

    real (single), pointer   :: tempImage(:, :)
    integer (int32), intent(in)    :: XSmoothSize, ySmoothSize

    ! internal variables
    integer (int32)                :: height
    integer (int32)                :: i

    ! For top hat
    real (single), pointer         :: scan(:)
    integer (int32), pointer       :: nscan(:)
    integer (int32)                :: allocErr

    tempImage => Allocate2DSingle(0, nx - 1, 0 , ny - 1, "tempImage")
     

    !*************************
    ! for gaussian smooth 
    !************************
    height = YSmoothSize
    
    scan => AllocateSingle(-height, height, "scan (smoothImage)")
    nscan => AllocateInt32(-height, height, "nscan (smoothImage)")
   

    !**************************************************
    ! smoothTopHat deals with one X position at a time
    !**************************************************
    do i = 0, nx - 1
        call MovingBox(image, tempImage, nx, ny, XSmoothSize, YSmoothSize, i, scan, nscan, 1)
    end do
  
    deallocate(scan, stat = allocErr)
    call CheckDeallocation(allocErr, "scan")

    deallocate(nscan, stat = allocErr)
    call CheckDeallocation(allocErr, "nscan")
    image = tempImage
    deallocate(tempImage)

!end subroutine SmoothImage1
end subroutine SmoothImage

!******************************************************************************
! subroutine MovingBox
! Top hat smoothing routine - does a strip in Y direction (best array access)
! 12.12.2001 - Modified to use the level image (ie to ignore non-background
! pixels)
!******************************************************************************
subroutine MovingBox(image, backgroundTot, nx, ny, width, height, startX, scan, &
                     nscan, type)

    integer (int32), intent(in)  :: nx, ny
    real (single), pointer    :: image(:, :)

   
    real (single), pointer :: backgroundTot(:, :)


    integer (int32), intent(in) :: width, height     ! Size of smoothing area
    integer (int32), intent(in) :: startX           ! starting point
    real (single), pointer :: scan(:)
    integer (int32), pointer :: nscan(:)

    integer (int32), intent(in) :: type
    !***********************
    ! internal variables
    ! for top hat smooth
    !***********************
    integer (int32)            :: i, j ,k
    real(single)               :: running_total
    integer (int32)            :: nrunning_total
    integer (int32)            :: minXpos, maxXpos
    integer (int32)            :: minYpos, maxYpos
    integer (int32)            :: clearStart, clearEnd
    integer (int32)            :: clearEndPlus1

    real (single)              :: scanval

    !************************************
    ! initialise arrays for this strip
    !************************************
    scan = 0
    nscan = 0

    ! X min and max's are fixed
     
    minXpos = MAX(startX - width, 0)
    maxXpos = MIN(startX + width, nx - 1)
 
    !**************************************************************
    ! Get start and end of clear (ie. don't need MIN/MAX checks )
    !**************************************************************
    clearStart = MIN(height + 1, ny)

    if(height .gt. ny) then
        clearEnd = -1
    else
        clearEnd = ny - height - 1
    end if

    !******************************************
    ! At start need to build up running total  
    ! First one
    !******************************************
    minYpos = 0
    maxYpos = MIN(height, ny - 1)


    running_total= 0.0
    nrunning_total= 0

    do j = minYpos, maxYpos     
        scanval = 0.0
        do k = minXpos, maxXpos 
            scanval = scanval + image(k, j)
        end do
        scan(j) = scanval
        nscan(j) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(j)

    end do
  
    if(type .eq. 1) then
        backgroundTot(startX, 0) = running_total / nrunning_total
    else
        backgroundTot(startX, 0) = running_total
    end if

  
    do i = 1, clearStart - 1
     
        ! Shift down arrays
     
        do j = minYpos,  maxYpos
            scan(j - 1) = scan(j)
            nscan(j - 1) = nscan(j)
        end do
     
        ! reset min/max values
     
        minYpos = MAX(i - height, 0)
        maxYpos = MIN(i + height, ny - 1)

        !***********************
        ! Add in new entry
        !***********************     
        scanval = 0.0
        do k = minXpos, maxXpos 
            scanval = scanval + image(k, maxYpos)
        end do
        minYpos = minYpos - i
        maxYpos = maxYpos - i
        scan(maxYpos) = scanval
        nscan(maxYpos) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(maxYpos)
 
        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    
    end do
 

    !********************************
    ! Now do the rest (not MIN/MAX)
    !********************************
    if(clearEnd .eq. -1) return
  

    ! Make sure we can hold whole smoothing area within data

    if(clearStart .lt. clearEnd) then
        do i = clearStart, clearEnd
     
            ! Shift down arrays and remove bottom line

            running_total = running_total - scan(-height)
            nrunning_total = nrunning_total - nscan(-height)
            do j = -height + 1, height
                scan(j - 1) = scan(j)
                nscan(j - 1) = nscan(j)
            end do
            !*******************
            ! Add in new entry
            !******************
        
            maxYpos = i + height
            scanval = 0.0
            do k = minXpos, maxXpos 
                scanval = scanval + image(k, maxYpos)
            end do

            scan(height) = scanval
            nscan(height) = maxXpos - minXpos + 1
            running_total = running_total + scanval
            nrunning_total = nrunning_total + nscan(height)
       
        
            if(type .eq. 1) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else
                backgroundTot(startX, i) = running_total
            end if

        end do
    end if

    !**********************
    ! And now do end part
    !**********************
    clearEndPlus1 = clearEnd + 1  

    do i = clearEndPlus1, ny - 1
        running_total = running_total - scan(height - i + clearEndPlus1)
        nrunning_total = nrunning_total - nscan(height - i + clearEndPlus1)

        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    end do

end subroutine MovingBox

!***************************************
! subroutine AddInt32Column
!***************************************
subroutine AddInt32ColumnG(tableH, value, title, units, columnDim) 
    type(tableT), intent(in) :: tableH
    integer(int32), intent(in) :: value
    character(*), intent(in) :: title
    character(*), intent(in) :: units
    integer(int32), intent(in), dimension(1) :: columnDim ! Column dimension.
    type(columnT) :: columnH
    integer(int32), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer32, units, columnDim)
    columnP => int32Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt32ColumnG 

end module omgprep_mod





























































































































































































































































